import ngsolve as ngs
from netgen.geom2d import SplineGeometry
import netgen.gui

geo = SplineGeometry()
geo.AddRectangle((0.0,0.0), (1.0,1.0), bc='exterior_bc')
ngmesh = geo.GenerateMesh(maxh=0.2, quad_dominated=True)
mesh = ngs.Mesh(ngmesh)

u_exact = 200 - ngs.x*ngs.x - 2*ngs.y*ngs.y
f = 6
g = 200 - ngs.x*ngs.x - 2*ngs.y*ngs.y

fes = ngs.H1(mesh, order=1, dirichlet='exterior_bc')
u = fes.TrialFunction()
v = fes.TestFunction()

a = ngs.BilinearForm(fes, symmetric=False)
a += ngs.SymbolicBFI(ngs.grad(u)*ngs.grad(v))

L = ngs.LinearForm(fes)
L += ngs.SymbolicLFI(f*v)

gfu = ngs.GridFunction(fes)
ngs.Draw(gfu)

gfu_exact = ngs.GridFunction(fes)

def SolveBVP():
    fes.Update()
    gfu.Update()
    a.Assemble()
    L.Assemble()
    
    gfu.Set(g, definedon=mesh.Boundaries('exterior_bc'))
    r = L.vec.CreateVector()
    r.data = L.vec - a.mat*gfu.vec
    gfu.vec.data += a.mat.Inverse(fes.FreeDofs())*r
    
    ngs.Redraw(blocking=True)

def CalcError():
    gfu_exact.Update()
    gfu_exact.Set(u_exact)
    
    err = (gfu - gfu_exact)*(gfu - gfu_exact)
    elerr = ngs.Integrate(err, mesh, ngs.VOL, element_wise=True)

    maxerr = max(elerr)
    print ("maxerr = ", maxerr)
    
    for el in mesh.Elements():
        mesh.SetRefinementFlag(el, elerr[el.nr] > 0.9*maxerr)
   
while fes.ndof < 1000:  
    SolveBVP()
    CalcError()
    mesh.Refine()
    
SolveBVP()